"""
Alice-Bob Integrator Environment (ABIE), created by Alice and Bob in the Moving Planets Around (MPA) project.

Features:
    1. The integrator takes input either from the commandline, or from the config file
    2. The integrator periodically stores data to HDF5
    3. The integrator supports restarting simulations


The MPA team, 2017-2018
"""
import argparse
import toml
from abie.integrator import Integrator
import numpy as np
import sys
from abie.data_io import DataIO
from abie.particles import Particles


class ABIE(object):
    def __init__(
        self,
        CONST_G=4 * np.pi ** 2,
        CONST_C=0.0,
        integrator="GaussRadau15",
        h=0.1,
        store_dt=100,
        name="simulation",
        buffer_len=10240,
        deviceID=-1
    ):
        # =================== CONSTANTS ==================
        # by default, using the square of the Gaussian gravitational constant
        self.__CONST_G = CONST_G
        self.__CONST_C = (
            CONST_C  # speed of light; PN terms will be calculated if CONST_C > 0
        )

        # # =================== VARIABLES ==================
        self.__t = 0.0
        self.__t_start = 0.0
        self.__t_end = 0.0
        self.__h = h
        self._particles = None
        self.__buf = None
        self.__store_dt = store_dt  # output is triggered per 100 time units
        self.__buffer_len = buffer_len  # size of dataset in a hdf5 group
        self.__max_close_encounter_events = 1
        self.__max_collision_events = 1
        self.__close_encounter_distance = 0.0
        self.__device_id = deviceID

        # load integrator modules
        self.__integrator_modules = (
            None  # a collection of integrators loaded from modules
        )
        self.__integrator_instances = dict()
        self.__integrator = None  # the actual, active integrator instance
        self.output_file = "%s.hdf5" % name
        self.__close_encounter_output_file = "%s_ce.txt" % name
        self.__collision_output_file = "%s_collisions.txt" % name

    @property
    def particles(self):
        if self._particles is None:
            self._particles = Particles(self.CONST_G)
        return self._particles

    @property
    def buffer(self):
        if self.__buf is None:
            self.__buf = DataIO(
                buf_len=self.__buffer_len,
                output_file_name=self.output_file,
                CONST_G=self.CONST_G,
            )
        return self.__buf

    @property
    def max_close_encounter_events(self):
        if self.__integrator is not None:
            self.__max_close_encounter_events = (
                self.__integrator.max_close_encounter_events
            )
            return self.__max_close_encounter_events
        else:
            return self.__max_close_encounter_events

    @max_close_encounter_events.setter
    def max_close_encounter_events(self, value):
        self.__max_close_encounter_events = value
        if self.__integrator is not None:
            self.__integrator.max_close_encounter_events = value

    @property
    def max_collision_events(self):
        if self.__integrator is not None:
            self.__max_collision_events = self.__integrator.collision_events
            return self.__max_collision_events
        else:
            return self.__max_collision_events

    @max_collision_events.setter
    def max_collision_events(self, value):
        self.__max_collision_events = value
        if self.__integrator is not None:
            self.__integrator.max_collision_events = value

    @property
    def close_encounter_distance(self):
        if self.__integrator is not None:
            self.__close_encounter_distance = (
                self.__integrator.max_close_encounter_events
            )
            return self.__close_encounter_distance
        else:
            return self.__close_encounter_distance

    @close_encounter_distance.setter
    def close_encounter_distance(self, value):
        self.__close_encounter_distance = value
        if self.__integrator is not None:
            self.__integrator.close_encounter_distance = value

    @property
    def close_encounter_output_file(self):
        if self.__integrator is not None:
            self.__close_encounter_output_file = (
                self.__integrator.close_encounter_output_file
            )
            return self.__close_encounter_output_file
        else:
            return self.__close_encounter_output_file

    @close_encounter_output_file.setter
    def close_encounter_output_file(self, value):
        self.__close_encounter_output_file = value
        if self.__integrator is not None:
            self.__integrator.close_encounter_output_file = value

    @property
    def collision_output_file(self):
        if self.__integrator is not None:
            self.__collision_output_file = self.__integrator.collision_output_file
            return self.__collision_output_file
        else:
            return self.__collision_output_file

    @collision_output_file.setter
    def collision_output_file(self, value):
        self.__collision_output_file = value
        if self.__integrator is not None:
            self.__integrator.collision_output_file = value

    @property
    def t(self):
        if self.__integrator is not None:
            self.__t = self.integrator.t
            return self.__t
        else:
            return self.__t

    @property
    def CONST_G(self):
        return self.__CONST_G

    @CONST_G.setter
    def CONST_G(self, value):
        self.__CONST_G = value
        if self.__integrator is not None:
            self.__integrator.CONST_G = value

    @property
    def CONST_C(self):
        return self.__CONST_C

    @CONST_C.setter
    def CONST_C(self, value):
        self.__CONST_C = value
        if self.__integrator is not None:
            self.__integrator.CONST_C = value

    @property
    def t_end(self):
        return self.__t_end

    @t_end.setter
    def t_end(self, tf):
        self.__t_end = tf
        if self.__integrator is not None:
            self.integrator.t_end = tf

    @property
    def h(self):
        if self.__integrator is not None:
            self.__h = self.integrator.h
            return self.__h
        else:
            return self.__h

    @h.setter
    def h(self, value):
        self.__h = value
        if self.__integrator is not None:
            self.__integrator.h = value

    @property
    def store_dt(self):
        if self.__integrator is not None:
            self.__store_dt = self.__integrator.__store_dt
            return self.__store_dt
        else:
            return self.__store_dt

    @store_dt.setter
    def store_dt(self, value):
        self.__store_dt = value
        if self.__integrator is not None:
            self.__integrator.store_dt = self.__store_dt

    @property
    def acceleration_method(self):
        return self.integrator.acceleration_method

    @acceleration_method.setter
    def acceleration_method(self, value):
        self.__integrator.acceleration_method = value

    @property
    def integrator(self):
        if self.__integrator is None:
            raise RuntimeError("Integrator not set!")
        return self.__integrator

    @integrator.setter
    def integrator(self, name_of_integrator):
        if self.__integrator_modules is None:
            self.__integrator_modules = Integrator.load_integrators()
        print(("Selecting %s as the active integrator." % name_of_integrator))
        # populate the parameters to the integrator
        if name_of_integrator in self.__integrator_modules:
            if name_of_integrator not in self.__integrator_instances:
                self.__integrator = getattr(
                    self.__integrator_modules[name_of_integrator], name_of_integrator
                )(self.particles, self.buffer, self.CONST_G, self.CONST_C, deviceID=self.__device_id)
                self.__integrator_instances[name_of_integrator] = self.__integrator
            else:
                self.__integrator = self.__integrator_instances[name_of_integrator]
            self.__integrator.CONST_G = self.CONST_G
            self.__integrator.t_end = self.__t_end
            self.__integrator.h = self.__h
            self.__integrator._t = self.__t
            self.__integrator.t_start = self.__t_start
            self.__integrator.output_file = self.output_file
            self.__integrator.collision_output_file = self.collision_output_file
            self.__integrator.close_encounter_output_file = (
                self.close_encounter_output_file
            )
        else:
            raise ValueError(
                "Unknown integrator %s. Supported integrators are %s"
                % (name_of_integrator, self.__integrator_modules)
            )

    @property
    def data(self):
        self.buffer.flush()
        return self.buffer.recorder.data

    def record_simulation(self, particles=None, quantities=None):
        if self.buffer.recorder is not None:
            self.buffer.recorder.set_monitored_particles(particles)
            self.buffer.recorder.set_monitored_quantities(quantities)
            return self.buffer.recorder
        else:
            from recorder import SimulationDataRecorder

            self.buffer.recorder = SimulationDataRecorder(particles, quantities)
            return self.buffer.recorder

    def initialize(self, config=None):
        # Initialize the integrator
        self.__integrator_modules = Integrator.load_integrators()
        if self.__integrator is None:
            print("Use GaussRadau15 as the default integrator...")
            self.integrator = "GaussRadau15"
            self.integrator.initialize()
            self.integrator.acceleration_method = "ctypes"
        else:
            self.__integrator.CONST_G = self.CONST_G
            self.__integrator.t_end = self.__t_end
            self.__integrator.h = self.__h
            self.__integrator.t_start = self.__t_start
            self.__integrator.output_file = self.output_file
            self.__integrator.store_dt = self.__store_dt
            self.__integrator.buffer_len = self.__buffer_len

        if config is not None:
            # Gravitational parameter
            self.integrator.CONST_G = np.array(config["physical_params"]["G"])

            # Integration parameters
            self.integrator = config["integration"]["integrator"]
            self.integrator.initialize()
            self.integrator.h = float(config["integration"]["h"])
            if "acc_method" in config["integration"]:
                self.integrator.acceleration_method = config["integration"][
                    "acc_method"
                ]
            else:
                self.integrator.acceleration_method = "ctypes"

            # Load sequence of object names
            if "names" in config:
                names = config["names"]
            else:
                names = None

            # Initial and final times
            if self.integrator.t_start == 0:
                self.integrator.t_start = float(config["integration"]["t0"])
            if self.integrator.t_end == 0:
                self.integrator.t_end = float(config["integration"]["tf"])
            self.integrator.active_integrator = config["integration"]["integrator"]
            DataIO.ic_populate(config["initial_conds"], self, names=names)

    def stop(self):
        """Stop the integrator and clean up the memory"""
        self.integrator.stop()

    def set_additional_forces(self, ext_acc):
        if ext_acc.ndim == 1 and ext_acc.shape[0] == 3 * self.integrator.particles.N:
            self.integrator.set_additional_forces(ext_acc)
        else:
            print(
                "WARNING: Additional forces array needs to be 3 * N vector, where N is the number of particles."
            )

    def integrate(self, to_time=None):
        try:
            ret = self.integrator.integrate(to_time)
            self.__t = self.integrator.t
        except KeyboardInterrupt as e:
            print(
                "Keyboard Interruption detected (Ctrl+C). Simulation stopped. Stopping the code..."
            )
            self.stop()
            sys.exit(0)

    def calculate_orbital_elements(self, primary=None):
        return self.integrator.calculate_orbital_elements(primary)

    def calculate_energy(self):
        return self.integrator.calculate_energy()

    def add(
        self,
        pos=None,
        vel=None,
        x=0.0,
        y=0.0,
        z=0.0,
        vx=0.0,
        vy=0.0,
        vz=0.0,
        mass=0.0,
        name=None,
        radius=0.0,
        ptype=0,
        a=None,
        e=0.0,
        i=0.0,
        Omega=0.0,
        omega=0.0,
        f=2 * np.pi * np.random.rand(),
        primary=None,
    ):
        if pos is None:
            pos = np.empty(3, dtype=np.double)
            pos[0] = x
            pos[1] = y
            pos[2] = z
        if vel is None:
            vel = np.empty(3, dtype=np.double)
            vel[0] = vx
            vel[1] = vy
            vel[2] = vz
        return self.particles.add(
            pos=pos,
            vel=vel,
            mass=mass,
            name=name,
            radius=radius,
            ptype=ptype,
            a=a,
            e=e,
            i=i,
            Omega=Omega,
            omega=omega,
            f=f,
            primary=primary,
        )


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-c", "--config", help="config file", default=None)
    parser.add_argument(
        "-o",
        "--output_file",
        dest="output_file",
        help="output data file",
        default="data.hdf5",
    )
    parser.add_argument(
        "-r", "--rebound_file", help="Rebound simulation file", default=None
    )
    parser.add_argument(
        "-t", "--t_end", type=float, dest="t_end", help="Termination time"
    )
    parser.add_argument(
        "-d",
        "--dt",
        type=float,
        dest="dt",
        help="Integration time step (optional for certain integrators)",
        default=None,
    )
    parser.add_argument(
        "-s",
        "--store_dt",
        type=float,
        dest="store_dt",
        help="output time step",
        default=100,
    )
    parser.add_argument(
        "-i",
        "--integrator",
        dest="integrator",
        help="Name of the integrator [GaussRadau15|WisdomHolman|RungeKutta|AdamsBashForth|LeapFrog|Euler]",
        default="GaussRadau15",
    )

    args = parser.parse_args()
    abie = ABIE()
    abie.integrator = args.integrator
    if args.output_file is not None:
        abie.output_file = args.output_file
    if args.t_end is not None:
        abie.t_end = args.t_end
    if args.config is not None:
        abie.initialize(DataIO.parse_config_file(args.config))
    elif args.rebound_file is not None:
        # populate the initial conditions from rebound simulation files
        abie.initialize()
        DataIO.ic_populate_from_rebound(args.rebound_file, abie)
    if args.dt is not None:
        abie.h = args.dt
    abie.store_dt = args.store_dt
    abie.integrate()


if __name__ == "__main__":
    main()



import numpy as np


class Tools(object):
    @staticmethod
    def compute_energy(sol_state, masses, const_g):
        n_rows, n_cols = sol_state.shape
        if n_cols % 6 == 0 and n_rows > 0:
            energy = np.zeros(n_rows)
            for t in range(n_rows):
                for i in range(0, n_cols / 6):
                    energy[t] += (
                        0.5
                        * masses[i]
                        * np.linalg.norm(
                            sol_state[t][n_cols / 2 + i * 3 : n_cols / 2 + i * 3 + 3]
                        )
                        ** 2
                    )
                    for j in range(0, n_cols / 6):
                        if i == j:
                            continue
                        energy[t] -= (
                            0.5
                            * const_g
                            * masses[i]
                            * masses[j]
                            / np.linalg.norm(
                                sol_state[t][i * 3 : i * 3 + 3]
                                - sol_state[t][j * 3 : j * 3 + 3]
                            )
                        )
            return energy

    @staticmethod
    def from_orbital_elements_to_cartesian(
        mp,
        ms,
        semimajor_axis,
        eccentricity,
        true_anomaly,
        inclination,
        argument_of_periapsis,
        longitude_of_ascending_node,
        G=4.0 * np.pi ** 2,
    ):
        """

        Function that returns position and velocities computed from the input orbital
        elements. Angles in radians, inclination between 0 and 180

        """

        cos_true_anomaly = np.cos(true_anomaly)
        sin_true_anomaly = np.sin(true_anomaly)

        cos_inclination = np.cos(inclination)
        sin_inclination = np.sin(inclination)

        cos_arg_per = np.cos(argument_of_periapsis)
        sin_arg_per = np.sin(argument_of_periapsis)

        cos_long_asc_nodes = np.cos(longitude_of_ascending_node)
        sin_long_asc_nodes = np.sin(longitude_of_ascending_node)

        ### e_vec is a unit vector directed towards periapsis ###
        e_vec_x = (
            cos_long_asc_nodes * cos_arg_per
            - sin_long_asc_nodes * sin_arg_per * cos_inclination
        )
        e_vec_y = (
            sin_long_asc_nodes * cos_arg_per
            + cos_long_asc_nodes * sin_arg_per * cos_inclination
        )
        e_vec_z = sin_arg_per * sin_inclination
        e_vec = np.array([e_vec_x, e_vec_y, e_vec_z])

        ### q is a unit vector perpendicular to e_vec and the orbital angular momentum vector ###
        q_vec_x = (
            -cos_long_asc_nodes * sin_arg_per
            - sin_long_asc_nodes * cos_arg_per * cos_inclination
        )
        q_vec_y = (
            -sin_long_asc_nodes * sin_arg_per
            + cos_long_asc_nodes * cos_arg_per * cos_inclination
        )
        q_vec_z = cos_arg_per * sin_inclination
        q_vec = np.array([q_vec_x, q_vec_y, q_vec_z])

        #    print 'alpha',alphax**2+alphay**2+alphaz**2 # For debugging; should be 1
        #    print 'beta',betax**2+betay**2+betaz**2 # For debugging; should be 1

        ### Relative position and velocity ###
        separation = (
            semimajor_axis
            * (1.0 - eccentricity ** 2)
            / (1.0 + eccentricity * cos_true_anomaly)
        )  # Compute the relative separation
        position_vector = (
            separation * cos_true_anomaly * e_vec
            + separation * sin_true_anomaly * q_vec
        )
        velocity_tilde = np.sqrt(
            G * (mp + ms) / (semimajor_axis * (1.0 - eccentricity ** 2))
        )  # Common factor
        velocity_vector = (
            -1.0 * velocity_tilde * sin_true_anomaly * e_vec
            + velocity_tilde * (eccentricity + cos_true_anomaly) * q_vec
        )

        return position_vector, velocity_vector

    @staticmethod
    def from_cartesian_to_orbital_elements(
        mp, ms, position, velocity, G=4.0 * np.pi ** 2
    ):
        """

        Function that computes orbital elements from cartesian coordinates.
        Return values are: mass1, mass2, semimajor axis, eccentricity,
        true anomaly, inclination, longitude of the ascending nodes and the
        argument of pericenter. All angles are returned in radians.
        In case of a perfectly circular orbit the true anomaly
        and argument of pericenter cannot be determined; in this case, the
        return values are 0.0 for both angles.

        """

        total_mass = mp + ms

        ### specific energy ###
        v_sq = np.dot(velocity, velocity)
        r_sq = np.dot(position, position)
        r = np.sqrt(r_sq)

        specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r
        # if specific_energy >= 0.0:
        #     print 'Not a bound orbit!'

        semimajor_axis = -G * total_mass / (2.0 * specific_energy)

        ### specific angular momentum ###
        specific_angular_momentum = np.cross(position, velocity)
        specific_angular_momentum_norm = np.sqrt(
            np.dot(specific_angular_momentum, specific_angular_momentum)
        )
        specific_angular_momentum_unit = (
            specific_angular_momentum / specific_angular_momentum_norm
        )

        maximum_specific_angular_momentum_norm = (
            G * total_mass / (np.sqrt(-2.0 * specific_energy))
        )
        ell = (
            specific_angular_momentum_norm / maximum_specific_angular_momentum_norm
        )  ### specific AM in units of maximum AM

        ### for e = 0 or e nearly 0, ell can be slightly larger than unity due to numerical reasons ###
        ell_epsilon = 1e-15

        completely_or_nearly_circular = False

        if ell > 1.0:
            if (
                1.0 < ell <= ell + ell_epsilon
            ):  ### still unity within numerical precision
                print(
                    "orbit is completely or nearly circular; in this case the LRL vector cannot be used to reliably obtain the argument of pericenter and true anomaly; the output values of the latter will be set to zero; output e will be e = 0"
                )
                ell = 1.0
                completely_or_nearly_circular = True
            else:  ### larger than unity within numerical precision
                raise Exception(
                    "angular momentum larger than maximum angular momentum for bound orbit"
                )

        eccentricity = np.sqrt(1.0 - ell ** 2)

        ### Orbital inclination ###
        z_vector = np.array([0.0, 0.0, 1.0])
        inclination = np.arccos(np.dot(z_vector, specific_angular_momentum_unit))

        ### Longitude of ascending nodes, with reference direction along x-axis ###
        ascending_node_vector = np.cross(z_vector, specific_angular_momentum)
        ascending_node_vector_norm = np.sqrt(
            np.dot(ascending_node_vector, ascending_node_vector)
        )
        if ascending_node_vector_norm == 0:
            ascending_node_vector_unit = np.array([1.0, 0.0, 0.0])
        else:
            ascending_node_vector_unit = (
                ascending_node_vector / ascending_node_vector_norm
            )

        long_asc_nodes = np.arctan2(
            ascending_node_vector_unit[1], ascending_node_vector_unit[0]
        )

        ### Argument of periapsis and true anomaly, using eccentricity a.k.a. Laplace-Runge-Lenz (LRL) vector ###
        mu = G * total_mass
        position_unit = position / r
        e_vector = (1.0 / mu) * np.cross(
            velocity, specific_angular_momentum
        ) - position_unit  ### Laplace-Runge-Lenz vector

        if (
            completely_or_nearly_circular == True
        ):  ### orbit is completely or nearly circular; in this case the LRL vector cannot be used to reliably obtain the argument of pericenter and true anomaly; the output values of the latter will be set to zero; output e will be e = 0
            arg_per = 0.0
            true_anomaly = 0.0
        else:
            e_vector_norm = np.sqrt(np.dot(e_vector, e_vector))
            e_vector_unit = e_vector / e_vector_norm

        e_vector_unit_cross_AM_unit = np.cross(
            e_vector_unit, specific_angular_momentum_unit
        )
        sin_arg_per = np.dot(ascending_node_vector_unit, e_vector_unit_cross_AM_unit)
        cos_arg_per = np.dot(e_vector_unit, ascending_node_vector_unit)
        arg_per = np.arctan2(sin_arg_per, cos_arg_per)

        sin_true_anomaly = np.dot(position_unit, -1.0 * e_vector_unit_cross_AM_unit)
        cos_true_anomaly = np.dot(position_unit, e_vector_unit)
        true_anomaly = np.arctan2(sin_true_anomaly, cos_true_anomaly)

        return (
            semimajor_axis,
            eccentricity,
            true_anomaly,
            inclination,
            arg_per,
            long_asc_nodes,
        )

    @staticmethod
    def from_cartesian_to_aei(
        mp,
        ms,
        position,
        velocity,
        G=0.0172020989 ** 2,
        z_vector=np.array([0.0, 0.0, 1.0]),
    ):
        """

        Function that computes orbital elements from cartesian coordinates.
        Return values are: mass1, mass2, semimajor axis, eccentricity,
        true anomaly, inclination, longitude of the ascending nodes and the
        argument of pericenter. All angles are returned in radians.
        In case of a perfectly circular orbit the true anomaly
        and argument of pericenter cannot be determined; in this case, the
        return values are 0.0 for both angles.

        """

        total_mass = mp + ms
        mu = G * total_mass

        if velocity.ndim == 1 and position.ndim == 1:
            # Calculate the a-e-i of one body wrt another one
            v_sq = np.dot(velocity, velocity)
            r_sq = np.dot(position, position)
            r = np.sqrt(r_sq)
            if r == 0:
                return np.nan, np.nan, np.nan

            specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r
            # if specific_energy >= 0.0:
            #     print 'Not a bound orbit!'

            semimajor_axis = -G * total_mass / (2.0 * specific_energy)

            specific_angular_momentum = np.cross(position, velocity)
            specific_angular_momentum_norm = np.sqrt(
                np.dot(specific_angular_momentum, specific_angular_momentum)
            )
            specific_angular_momentum_unit = (
                specific_angular_momentum / specific_angular_momentum_norm
            )
            ecc_vec = (
                (v_sq - mu / r) * position - np.dot(position, velocity) * velocity
            ) / mu

            inclination = np.arccos(np.dot(z_vector, specific_angular_momentum_unit))
            eccentricity = np.linalg.norm(ecc_vec)

        elif velocity.ndim == 2 and position.ndim == 2:

            # calculate a list of bodies (pos and vel are n*3 arrays, representing n points)
            v_sq = np.sum(velocity * velocity, axis=0)
            r_sq = np.sum(position * position, axis=0)
            r = np.sqrt(r_sq)

            specific_energy = (1.0 / 2.0) * v_sq - G * total_mass / r

            semimajor_axis = -G * total_mass / (2.0 * specific_energy)

            ### specific angular momentum ###
            specific_angular_momentum = np.cross(
                position.transpose(), velocity.transpose()
            ).transpose()
            specific_angular_momentum_norm = np.sqrt(
                np.sum(specific_angular_momentum * specific_angular_momentum, axis=0)
            )
            specific_angular_momentum_unit = (
                specific_angular_momentum / specific_angular_momentum_norm
            )
            ecc_vec = (
                (v_sq - mu / r) * position
                - np.sum(position * velocity, axis=0) * velocity
            ) / mu

            eccentricity = np.linalg.norm(ecc_vec, axis=0)

            ### Orbital inclination ###
            inclination = np.arccos(np.matmul(z_vector, specific_angular_momentum_unit))

        return semimajor_axis, eccentricity, inclination



from setuptools import setup, find_packages

try:
    from setuptools import setup, Extension
except ImportError:
    from distutils.core import setup, Extension

from distutils.command.build_ext import build_ext
import sys
import sysconfig
import os


def locate_cuda():
    if "CUDAHOME" in os.environ:
        home = os.environ["CUDAHOME"]
        nvcc = os.path.join(home, "bin", "nvcc")
        cudaconfig = {
            "home": home,
            "nvcc": nvcc,
            "include": os.path.join(home, "include"),
            "lib64": os.path.join(home, "lib64"),
        }
        print("Found CUDA with the following config: ", cudaconfig)
        return cudaconfig
    else:
        for directory in os.environ["PATH"].split(os.pathsep):
            nvcc = os.path.join(directory, "nvcc")
            if os.path.exists(nvcc):
                # further check whether the header files and library are there
                home = os.path.dirname(os.path.dirname(nvcc))
                if os.path.exists(os.path.join(home, "include")) and os.path.exists(
                    os.path.join(home, "lib64")
                ):
                    cudaconfig = {
                        "home": home,
                        "nvcc": nvcc,
                        "include": os.path.join(home, "include"),
                        "lib64": os.path.join(home, "lib64"),
                    }
                    print("Found CUDA with the following config: ", cudaconfig)
                    return cudaconfig
        return None


def customize_compiler_for_nvcc(self):
    """
    inject deep into distutils to customize how the dispatch
    to gcc/nvcc works.
    If you subclass UnixCCompiler, it's not trivial to get your subclass
    injected in, and still have the right customizations (i.e.
    distutils.sysconfig.customize_compiler) run on it. So instead of going
    the OO route, I have this. Note, it's kindof like a wierd functional
    subclassing going on.
    """

    # tell the compiler it can processes .cu
    self.src_extensions.append(".cu")

    # save references to the default compiler_so and _comple methods
    default_compiler_so = self.compiler_so
    super = self._compile

    # now redefine the _compile method. This gets executed for each
    # object but distutils doesn't have the ability to change compilers
    # based on source extension: we add it.
    def _compile(obj, src, ext, cc_args, extra_postargs, pp_opts):
        if os.path.splitext(src)[1] == ".cu":
            # use the cuda for .cu files
            self.set_executable("compiler_so", CUDA["nvcc"])
            # use only a subset of the extra_postargs, which are 1-1 translated
            # from the extra_compile_args in the Extension class
            postargs = extra_postargs["nvcc"]
        else:
            postargs = extra_postargs["gcc"]

        super(obj, src, ext, cc_args, postargs, pp_opts)
        # reset the default compiler_so, which we might have changed for cuda
        self.compiler_so = default_compiler_so

    # inject our redefined _compile method into the class
    self._compile = _compile


# run the customize_compiler
class cuda_build_ext(build_ext):
    def build_extensions(self):
        customize_compiler_for_nvcc(self.compiler)
        build_ext.build_extensions(self)


def find_files_with_ext(path, ext):
    files_list = list()
    for root, dirs, files in os.walk(path):
        for name in files:
            try:
                current_ext = name.split(".")[-1]
            except IndexError:
                continue

            if current_ext == ext:
                files_list.append(join(root, name))
    return files_list


suffix = sysconfig.get_config_var("EXT_SUFFIX")
if suffix is None:
    suffix = ".so"

extra_link_args = []
if sys.platform == "darwin":
    from distutils import sysconfig

    vars = sysconfig.get_config_vars()
    vars["LDSHARED"] = vars["LDSHARED"].replace("-bundle", "-shared")
    extra_link_args = ["-Wl,-install_name,@rpath/libabie" + suffix]

sources = [
    "src/integrator_gauss_radau15.c",
    "src/integrator_wisdom_holman.c",
    "src/integrator_runge_kutta.c",
    "src/common.c",
    "src/additional_forces.c",
]

# Locate CUDA paths
opt_cuda = os.getenv("WITH_CUDA")
if opt_cuda is None or opt_cuda == str(1):
    CUDA = locate_cuda()
else:
    CUDA = None

gcc_compile_flags = [
    "-fstrict-aliasing",
    "-O3",
    "-std=c99",
    "-fPIC",
    "-shared",
    "-fcommon",
    "-fcommon",
]
if CUDA is not None:
    # CUDA/GPU detected on the system
    sources.append("src/gpuforce.cu")
    sources.append("src/bodysystemcuda.cu")
    extra_link_args = ["-fPIC", "-DGPU", "-O3", "-g", "-lcudart", "-lstdc++"]
    gcc_compile_flags.append("-DGPU")
    print("CUDA SDK found! ABIE will be built with GPU support.")
else:
    print("CUDA SDK  not found! ABIE will be built without GPU support.")

module_abie = Extension(
    name="libabie",
    sources=sources,
    extra_link_args=extra_link_args,
    extra_compile_args={
        "gcc": gcc_compile_flags,
        "nvcc": [
            "-O3",
            "-Xcompiler",
            "-DGPU",
            "--use_fast_math",
            "--ptxas-options=-v",
            "-c",
            "--compiler-options",
            "-fPIC",
        ],
    },
)

setup(
    name="abie",
    version="0.8.5",
    description="Alice-Bob Integrator Environment (ABIE), a GPU-accelerated integrator framework for astrophysical N-body simulations",
    url="https://github.com/MovingPlanetsAround/ABIE",
    author="Maxwell X. Cai, Javier Roa, Adrian S. Hamers, Nathan W. C. Leigh",
    author_email="maxwellemail@gmail.com",
    license="GPL-3.0",
    packages=find_packages(),
    zip_safe=False,
    install_requires=["toml", "numpy", "h5py", "six"],
    entry_points={"console_scripts": ["abie = ABIE.abie:main"]},
    cmdclass={
        "build_ext": cuda_build_ext,
    },
    ext_modules=[module_abie],
)
